from typing import Any, Dict, List, Tuple

import networkx as nx
import numpy as np
from spacy.tokens.doc import Doc

from spacy_ke.base import KeywordExtractor
from spacy_ke.util import Candidate


class TextRank(KeywordExtractor):
    """spaCy implementation of "TextRank: Bringing Order into Text".

    Usage example:
    --------------

    >>> import spacy

    >>> nlp = spacy.load("en_core_web_sm")
    >>> nlp.add_pipe(TextRank(nlp))

    >>> doc = nlp(
        "Natural language processing (NLP) is a subfield of linguistics, computer science, and artificial intelligence "
        "concerned with the interactions between computers and human language, in particular how to program computers "
        "to process and analyze large amounts of natural language data. "
    )

    >>> doc._.extract_keywords(n=5)
    """

    defaults: Dict[str, Any] = {
        "pos": frozenset({"ADJ", "NOUN", "PROPN", "VERB"}),
        "window": 3,
        "alpha": 0.85,
        "tol": 1.0e-6,
        "candidate_selection": "chunk",
    }

    def candidate_weighting(self, doc: Doc) -> List[Tuple[Candidate, float]]:
        """Compute the weighted score of each keyword candidate.

        Args:
            doc (Doc): doc.

        Returns:
            list of tuples, candidate with a score.
        """
        res = []
        G = self.build_graph(doc)
        W = nx.pagerank_scipy(G, alpha=self.cfg["alpha"], tol=self.cfg["tol"])
        for candidate in doc._.kw_candidates:
            chunk_len = len(candidate.lexical_form)
            non_lemma = 0
            rank = 0.0
            for t in candidate.lexical_form:
                if t in W:
                    rank += W[t]
                else:
                    non_lemma += 1
            non_lemma_discount = chunk_len / (chunk_len + (2.0 * non_lemma) + 1.0)
            candidate_w = np.sqrt(rank / (chunk_len + non_lemma)) * non_lemma_discount
            candidate_w += (
                candidate.offsets[0] * 1e-8
            )  # break ties according to position in text
            res.append((candidate, candidate_w))
        res.sort(key=lambda x: x[1], reverse=True)
        return res

    def build_graph(self, doc: Doc):
        """Build a lemma graph representation of a document.

        Args:
            doc (Doc): doc.

        Returns:
            nx.Graph
        """
        G = nx.Graph()
        pos = self.cfg["pos"]
        window_size = self.cfg["window"]
        seen = set()
        for sent in doc.sents:
            for token in sent:
                if token.is_stop or token.pos_ not in pos:
                    continue
                node0 = token.lemma_.lower()
                if not G.has_node(node0):
                    G.add_node(node0)
                for prev_token in sent[max(sent.start, token.i - window_size) : token.i]:
                    node1 = prev_token.lemma_.lower()
                    if node0 != node1 and node1 in seen:
                        if G.has_edge(node0, node1):
                            G[node0][node1]["weight"] += 1
                        else:
                            G.add_edge(node0, node1, weight=1)
                seen.add(node0)
        return G



import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Tuple

import numpy as np
from spacy.tokens.doc import Doc

from spacy_ke.base import KeywordExtractor
from spacy_ke.util import Candidate


@dataclass
class YakeFeatures:
    """ term frequency """

    tf: float

    """
    1. CASING: gives importance to acronyms or words starting with a capital letter.
               CASING(w) = max(TF(U(w)), TF(A(w))) / (1 + log(TF(w)))
               with TF(U(w) being the # times the word starts with an uppercase
               letter, excepts beginning of sentences. TF(A(w)) is the # times
               the word is marked as an acronym.
    """
    casing: float

    """
    2. POSITION: gives importance to words occurring at the beginning of the document.
               POSITION(w) = log( log( 3 + Median(Sen(w)) ) )
               with Sen(w) contains the position of the sentences where w
               occurs.
    """

    position: float

    """
    3. FREQUENCY: gives importance to frequent words.
               FREQUENCY(w) = TF(w) / ( MEAN_TF + STD_TF)
               with MEAN_TF and STD_TF computed on valid_tfs which are words
               that are not stopwords.
    """
    frequency: float

    """
    4. RELATEDNESS: gives importance to words that do not have the characteristics of stopwords.
               RELATEDNESS(w) = 1 + (WR+WL)*(TF(w)/MAX_TF) + PL + PR
    """
    relatedness: float

    """
    5. DIFFERENT: gives importance to words that occurs in multiple sentences.
               DIFFERENT(w) = SF(w) / # sentences
               with SF(w) being the sentence frequency of word w.
    """
    different: float

    """ right and left context words """
    ctx: Tuple[List[str], List[str]]

    @property
    def weight(self):
        A = self.casing
        B = self.position
        C = self.frequency
        D = self.relatedness
        E = self.different

        return (D * B) / (A + (C / D) + (E / D))


class Yake(KeywordExtractor):
    """spaCy implementation of "YAKE! Collection-Independent Automatic Keyword Extractor".

    Usage example:
    --------------

    >>> import spacy

    >>> nlp = spacy.load("en_core_web_sm")
    >>> nlp.add_pipe(Yake(nlp))

    >>> doc = nlp(
        "Natural language processing (NLP) is a subfield of linguistics, computer science, and artificial intelligence "
        "concerned with the interactions between computers and human language, in particular how to program computers "
        "to process and analyze large amounts of natural language data. "
    )

    >>> doc._.extract_keywords(n=5)
    """

    defaults: Dict[str, Any] = {
        "window": 2,
        "lemmatize": False,
        "candidate_selection": "ngram",
    }

    def candidate_weighting(self, doc: Doc) -> List[Tuple[Candidate, Any]]:
        """Compute the weighted score of each keyword candidate.

        Args:
            doc (Doc): doc.

        Returns:
            list of tuples, candidate with a score.
        """
        res = []
        vocab = self._build_vocab_features(doc)
        for candidate in doc._.kw_candidates:
            if self.cfg["lemmatize"]:
                n_offsets = len(candidate.offsets)
                weights = [vocab[w].weight for w in candidate.lexical_form]
                candidate_w = np.prod(weights) / (n_offsets * (1 + sum(weights)))
                res.append((candidate, candidate_w))
            else:
                lowercase_forms = [
                    " ".join(t.text.lower() for t in sf) for sf in candidate.surface_forms
                ]
                for i, sf in enumerate(candidate.surface_forms):
                    tf = lowercase_forms.count(sf.text.lower())
                    prod_ = 1.0
                    sum_ = 0.0
                    for j, token in enumerate(sf):
                        if token.is_stop:
                            term_stop = token.lower_
                            prob_t1 = prob_t2 = 0
                            if j - 1 >= 0:
                                left = sf[j - 1]
                                prob_t1 = (
                                    vocab[left.lower_].ctx[1].count(term_stop)
                                    / vocab[left.lower_].tf
                                )
                            if j + 1 < len(sf):
                                right = sf[j + 1]
                                prob_t2 = (
                                    vocab[term_stop].ctx[0].count(right.text)
                                    / vocab[right.lower_].tf
                                )
                            prob = prob_t1 * prob_t2
                            prod_ *= 1 + (1 - prob)
                            sum_ -= 1 - prob
                        else:
                            prod_ *= vocab[token.lower_].weight
                            sum_ += vocab[token.lower_].weight
                    if sum_ == -1:
                        # The candidate is a one token stopword at the start or
                        #  the end of the sentence
                        # Setting sum_ to -1+eps so 1+sum_ != 0
                        sum_ = -0.99999999999
                    candidate_w = prod_
                    candidate_w /= tf * (1 + sum_)
                    res.append((candidate, candidate_w))
        res.sort(key=lambda x: x[1])
        return res

    def _build_vocab_features(self, doc: Doc) -> Dict[str, YakeFeatures]:
        """Compute the weight of individual words using yake features (see ``Features`` class).

        Args:
            doc (Doc): doc.

        Returns:
            dict[word<str> -> features<Features>]
        """
        features = dict()
        vocab = self._build_vocab(doc, lemmatize=self.cfg["lemmatize"])
        contexts = self._build_ctx(doc, vocab, window=self.cfg["window"])
        n_sentences = len(list(doc.sents))
        stop_words = self.nlp.Defaults.stop_words
        freqs = [len(vocab[w]) for w in vocab]
        freqs_nsw = [len(vocab[w]) for w in vocab if stop_words and w not in stop_words]
        mean_freq = np.mean(freqs_nsw)
        std_freq = np.std(freqs_nsw)
        max_freq = np.max(freqs)

        for word in vocab:
            # Term Frequency
            #   & Uppercase/Acronym Term Frequencies
            tf = len(vocab[word])
            tf_a = 0
            tf_u = 0

            for (offset, shift, sent_id, surface_form) in vocab[word]:
                if surface_form.isupper() and len(word) > 1:
                    tf_a += 1
                elif surface_form[0].isupper() and offset != shift:
                    tf_u += 1

            # 1. CASING
            casing = max(tf_a, tf_u) / (1.0 + np.log(tf))

            # 2. POSITION
            sentence_ids = list(set([t[2] for t in vocab[word]]))
            position = np.log(3.0 + np.median(sentence_ids))
            position = np.log(position)

            # 3. FREQUENCY
            frequency = tf
            frequency /= mean_freq + std_freq

            # 4. RELATEDNESS
            wl = 0.0
            if len(contexts[word][0]):
                wl = len(set(contexts[word][0]))
                wl /= len(contexts[word][0])
            pl = len(set(contexts[word][0])) / max_freq

            wr = 0.0
            if len(contexts[word][1]):
                wr = len(set(contexts[word][1]))
                wr /= len(contexts[word][1])
            pr = len(set(contexts[word][1])) / max_freq

            relatedness = 1
            relatedness += (wr + wl) * (tf / max_freq)
            relatedness += pl
            relatedness += pr

            # 5. DIFFERENT
            different = len(sentence_ids) / n_sentences
            features[word] = YakeFeatures(
                tf, casing, position, frequency, relatedness, different, contexts[word]
            )

        return features

    @staticmethod
    def _build_vocab(doc: Doc, lemmatize: bool = False) -> Dict[str, Tuple[int, int, str]]:
        """Build the vocabulary that will be used to weight candidates.
        Only words containing at least one alpha-numeric character are kept.

        Args:
            doc (Doc): doc.
            lemmatize (bool): whether or not to lemmatize vocabulary.

        Returns:
            dict.
        """
        words = defaultdict(set)
        for sent_i, sent in enumerate(doc.sents):
            for token in sent:
                if token.is_alpha and not re.search("(?i)^-[lr][rcs]b-$", token.text):
                    if lemmatize:
                        word = token.lemma_.lower()
                    else:
                        word = token.lower_
                    words[word].add((sent.start + token.i, sent.start, sent_i, token.text))
        return words

    @staticmethod
    def _build_ctx(doc: Doc, vocab: Dict, window: int = 2) -> Dict[str, Tuple[List, List]]:
        """Build the contexts of the words for computing the relatedness
        feature. Words that occur within a window of n words are considered as
        context words. Only words co-occurring in a block (sequence of words
        that appear in the vocabulary) are considered.

        Args:
            doc (Doc): doc.
            vocab (Dict): See ``_vocab()`` method.
            window (int): optional, size of context window, default is 2.

        Returns:
            dict.
        """
        contexts = defaultdict(lambda: ([], []))
        for sent_i, sent in enumerate(doc.sents):
            block = []
            for token in sent:
                word = token.lower_
                if word not in vocab:
                    block = []
                    continue
                # add the left context
                contexts[word][0].extend(
                    [w for w in block[max(0, len(block) - window) : len(block)]]
                )
                # add the right context
                for w in block[max(0, len(block) - window) : len(block)]:
                    contexts[w][1].append(word)
                # add word to the current block
                block.append(word)
        return contexts



import itertools
from typing import Any, Dict, Iterable, List, Tuple

import networkx as nx
import numpy as np
from scipy.cluster.hierarchy import fcluster, linkage
from scipy.spatial.distance import pdist
from spacy.tokens.doc import Doc

from spacy_ke.base import KeywordExtractor
from spacy_ke.util import Candidate


class TopicRank(KeywordExtractor):
    """spaCy implementation of "TopicRank: Graph-Based Topic Ranking for Keyphrase Extraction".

    Usage example:
    --------------

    >>> import spacy

    >>> nlp = spacy.load("en_core_web_sm")
    >>> nlp.add_pipe(TopicRank(nlp))

    >>> doc = nlp(
        "Natural language processing (NLP) is a subfield of linguistics, computer science, and artificial intelligence "
        "concerned with the interactions between computers and human language, in particular how to program computers "
        "to process and analyze large amounts of natural language data. "
    )

    >>> doc._.extract_keywords(n=5)
    """

    defaults: Dict[str, Any] = {
        "clustering_method": "average",
        "distance_metric": "jaccard",
        "threshold": 0.74,
        "alpha": 0.85,
        "tol": 1.0e-6,
        "heuristic": None,
        "candidate_selection": "chunk",
    }

    def candidate_selection(self, doc: Doc) -> Iterable[Candidate]:
        """Get keywords candidates.

        Args:
            doc (Doc): doc.

        Returns:
            Iterable[Candidate]
        """
        C = super().candidate_selection(doc)
        C.sort(key=lambda x: x.lexical_form)
        return C

    def candidate_weighting(self, doc: Doc) -> List[Tuple[Candidate, float]]:
        """Compute the weighted score of each keyword candidate.

        Args:
            doc (Doc): doc.

        Returns:
            list of tuples, candidate with a score.
        """
        res = []
        C = doc._.kw_candidates
        G = self.build_graph(doc)
        W = nx.pagerank_scipy(
            G, alpha=self.cfg["alpha"], tol=self.cfg["tol"], weight="weight"
        )

        for i, topic in nx.get_node_attributes(G, "C").items():
            offsets = [C[t].offsets[0] for t in topic]
            if self.cfg["heuristic"] == "frequent":
                freq = [len(C[t].surface_forms) for t in topic]
                indexes = [j for j, f in enumerate(freq) if f == max(freq)]
                indexes_offsets = [offsets[j] for j in indexes]
                most_frequent = offsets.index(min(indexes_offsets))
                res.append((C[topic[most_frequent]], W[i]))
            else:
                first = offsets.index(min(offsets))
                res.append((C[topic[first]], W[i]))

        res.sort(key=lambda x: x[1], reverse=True)
        return res

    def build_graph(self, doc: Doc) -> nx.Graph:
        """Build topic graph from candidates.

        Args:
            doc (Doc): doc.

        Returns:
            nx.Graph
        """
        G = nx.Graph()
        C = doc._.kw_candidates
        T = self.topic_clustering(
            doc,
            clustering_method=self.cfg["clustering_method"],
            distance_metric=self.cfg["distance_metric"],
            threshold=self.cfg["threshold"],
        )
        n_topics = len(T)
        for i, j in itertools.combinations(range(n_topics), 2):
            if not G.has_node(i):
                G.add_node(i, C=T[i])
            if not G.has_node(j):
                G.add_node(j, C=T[j])
            G.add_edge(i, j, weight=0.0)
            for c_i in T[i]:
                for c_j in T[j]:
                    for p_i in C[c_i].offsets:
                        for p_j in C[c_j].offsets:
                            gap = abs(p_i - p_j)
                            if p_i < p_j:
                                gap -= len(C[c_i].lexical_form) - 1
                            if p_j < p_i:
                                gap -= len(C[c_j].lexical_form) - 1
                            G[i][j]["weight"] += 1.0 / gap
        return G

    def topic_clustering(
        self,
        doc: Doc,
        clustering_method="average",
        distance_metric="jaccard",
        threshold=0.74,
    ) -> List[List[int]]:
        """Get a list of topics by clustering candidates' lexical forms.

        Args:
            doc (Doc): doc.
            clustering_method (str): optional, clustering method for linkage. default is "average".
            distance_metric (str): optional, distance metric for similarity. default is "jaccard".
            threshold (float): the minimum similarity for clustering, defaults to 0.74,
                               i.e. more than 1/4 of lexical form overlap similarity.

        Returns:
            List of topics, each represented by a list of candidate indices.
        """
        topics = []
        C = doc._.kw_candidates
        if len(C) == 1:
            topics.append([0])
        else:
            X = self.embed(C)
            Y = pdist(X, distance_metric)
            Z = linkage(Y, method=clustering_method)
            clusters = fcluster(Z, t=threshold, criterion="distance")
            for cluster_id in range(1, max(clusters) + 1):
                topics.append(
                    [j for j in range(len(clusters)) if clusters[j] == cluster_id]
                )
        return topics

    @staticmethod
    def embed(candidates: List[Candidate]) -> np.ndarray:
        """Get embedding matrix of candidates.

        Args:
            candidates (list): List of candidates.

        Returns:
            np.ndarray.
        """
        dim = set()
        for candidate in candidates:
            for w in candidate.lexical_form:
                dim.add(w)
        dim = list(dim)
        X = np.zeros((len(candidates), len(dim)))
        for i, c in enumerate(candidates):
            for w in c.lexical_form:
                X[i, dim.index(w)] += 1
        return X


if __name__ == "__main__":
    import spacy

    nlp = spacy.load("en_core_web_sm")
    nlp.add_pipe(TopicRank(nlp))

    doc = nlp(
        "Natural language processing (NLP) is a subfield of linguistics, computer science, and artificial intelligence "
        "concerned with the interactions between computers and human language, in particular how to program computers "
        "to process and analyze large amounts of natural language data. "
    )

    for keyword, score in doc._.extract_keywords(n=10):
        print(keyword, "-", score)



from typing import Any, Dict, Iterable, List, Tuple

from spacy import displacy
from spacy.language import Language
from spacy.tokens.doc import Doc

from spacy_ke.util import Candidate, registry


class KeywordExtractor:
    defaults: Dict[str, Any] = {"candidate_selection": "ngram"}

    def __init__(self, nlp: Language, **overrides):
        self.nlp = nlp
        self.cfg = self.defaults.copy()
        self.cfg.update(overrides)

    def __call__(self, doc: Doc) -> Doc:
        self.init_component()
        doc._.kw_candidates = self.candidate_selection(doc)
        return doc

    def init_component(self):
        if not Doc.has_extension("extract_keywords"):
            Doc.set_extension("extract_keywords", method=self.extract_keywords)
        if not Doc.has_extension("kw_candidates"):
            Doc.set_extension("kw_candidates", default=None)

    def render(self, doc: Doc, jupyter=None, **kw_kwargs):
        """Render HTML for text highlighting of keywords.

        Args:
            doc (Doc): doc.
            jupyter (bool): optional, override jupyter auto-detection.
            kw_kwargs (kwargs): optional, keyword arguments for ``kw``.

        Returns:
            Rendered HTML markup
        """
        doc = self(doc)
        spans = doc._.extract_keywords(**kw_kwargs)
        examples = [
            {
                "text": doc.text,
                "title": None,
                "ents": sorted(
                    [
                        {
                            "start": span.start_char,
                            "end": span.end_char,
                            "label": f"{round(score, 3)}",
                        }
                        for (span, score) in spans
                    ],
                    key=lambda x: x["start"],
                ),
            }
        ]
        html = displacy.render(examples, style="ent", manual=True, jupyter=jupyter)
        return html

    def extract_keywords(self, doc: Doc, n=10, similarity_thresh=0.45):
        """Returns the n-best candidates given the weights.

        Args:
            doc (Doc): doc.
            n (int): the number of candidates, defaults to 10.
            similarity_thresh (float): optional, similarity thresh for dedup. default is 0.45.

        Returns:
            List
        """
        spans = []
        candidates_weighted = self.candidate_weighting(doc)
        for candidate, candidate_w in candidates_weighted:
            if similarity_thresh > 0.0:
                redundant = False
                for prev_candidate, _ in spans:
                    if candidate.similarity(prev_candidate) > similarity_thresh:
                        redundant = True
                        break
                if redundant:
                    continue
            spans.append((candidate, candidate_w))
            if len(spans) >= n:
                break
        spans = spans[: min(n, len(spans))]
        spans = [(c.surface_forms[0], score) for c, score in spans]
        return spans

    def candidate_weighting(self, doc: Doc) -> List[Tuple[Candidate, Any]]:
        """Compute the weighted score of each keyword candidate.

        Args:
            doc (Doc): doc.

        Returns:
            list of tuples, candidate with a score.
        """
        return [(c, 1.0) for c in doc._.kw_candidates]

    def candidate_selection(self, doc: Doc) -> Iterable[Candidate]:
        """Get keywords candidates.

        Args:
            doc (Doc): doc.

        Returns:
            Iterable[Candidate]
        """
        candidate_selection = registry.candidate_selection.get(
            self.cfg["candidate_selection"]
        )
        return candidate_selection(doc)



__title__ = "spacy_ke"
__version__ = "0.1.4"
__summary__ = "Keyword extraction with spaCy"
__author__ = "Tal Almagor"
__email__ = "almagoric@gmail.com"
__license__ = "GPL"
__url__ = "https://github.com/talmago/spacy_ke"



from dataclasses import dataclass
from typing import Iterable, Iterator, List, Set, Tuple

import catalogue
import editdistance
from spacy.tokens.doc import Doc
from spacy.tokens.span import Span
from spacy.util import filter_spans


@dataclass
class Candidate:
    lexical_form: List[str]
    """ the lexical form of the candidate. """

    surface_forms: List[Span]
    """ the surface forms of the candidate. """

    sentence_ids: List[int]
    """ the sentence id of each surface form. """

    @property
    def offsets(self):
        return [sf.start for sf in self.surface_forms]

    def similarity(self, other):
        """Measure a similarity to another candidate.

        ``SIM = 1 - edit_distance(lexical_form1, lexical_form2) / max(#lexical_form1, #lexical_form2)``.
        """
        assert isinstance(other, Candidate)

        lf1 = " ".join(self.lexical_form)
        lf2 = " ".join(other.lexical_form)
        dist = editdistance.eval(lf1, lf2)
        dist /= max(len(lf1), len(lf2))
        return 1.0 - dist


class registry(object):
    candidate_selection = catalogue.create("spacy_ke", "candidate_selection")


@registry.candidate_selection.register("chunk")
def chunk_selection(doc: Doc) -> Iterable[Candidate]:
    """Get keywords candidates from noun chunks and entities.

    Args:
        doc (Doc): doc.

    Returns:
        Iterable[Candidate]
    """
    surface_forms = []
    spans = list(doc.ents)
    ent_words: Set[str] = set()
    sentence_indices = []
    for span in spans:
        ent_words.update(token.i for token in span)
    for np in doc.noun_chunks:
        # https://github.com/explosion/sense2vec/blob/c22078c4e6c13038ab1c7718849ff97aa54fb9d8/sense2vec/util.py#L105
        while len(np) > 1 and np[0].dep_ not in ("advmod", "amod", "compound"):
            np = np[1:]
        if not any(w.i in ent_words for w in np):
            spans.append(np)
    for sent in doc.sents:
        sentence_indices.append((sent.start, sent.end))
    for span in filter_spans(spans):
        for i, token_indices in enumerate(sentence_indices):
            if span.start >= token_indices[0] and span.end <= token_indices[1]:
                surface_forms.append((i, span))
                break
    return _merge_surface_forms(surface_forms)


@registry.candidate_selection.register("ngram")
def ngram_selection(doc: Doc, n=3) -> Iterable[Candidate]:
    """Get keywords candidates from ngrams.

    Args:
        doc (Doc): doc.
        n (int): ngram range.

    Returns:
        Iterable[Candidate]
    """

    def _is_candidate(span: Span, min_length=3, min_word_length=2, alpha=True) -> bool:
        """Check if N-gram span is qualified as a candidate.

        Args:
            span (Span): n-gram.
            min_length (int): minimum length for n-gram.
            min_word_length (int): minimum length for word in an ngram.
            alpha (bool): Filter n-grams with non-alphanumeric words.

        Returns:
            bool
        """
        n_span = len(span)
        # discard if composed of 1-2 characters
        if len(span.text) < min_length:
            return False
        for token_idx, token in enumerate(span):
            # discard if contains punct
            if token.is_punct:
                return False
            # discard if contains tokens of 1-2 characters
            if len(token.text) < min_word_length:
                return False
            # discard if contains non alphanumeric
            if alpha and not token.is_alpha:
                return False
            # discard if first/last word is a stop
            if token_idx in (0, n_span - 1) and token.is_stop:
                return False
        # discard if ends with NOUN -> VERB
        if n_span >= 2 and span[-2].pos_ == "NOUN" and span[-1].pos_ == "VERB":
            return False
        return True

    surface_forms = [sf for sf in _ngrams(doc, n=n) if _is_candidate(sf[1])]
    return _merge_surface_forms(surface_forms)


def _ngrams(doc: Doc, n=3) -> Iterator[Tuple[int, Span]]:
    """Select all the n-grams and populate the candidate container.

    Args:
        doc (Doc): doc.
        n (int): the n-gram length, defaults to 3.

    Returns:
        Iterator(sentence_id<int>, offset<int>, ngram<Span>)
    """
    for sentence_id, sentence in enumerate(doc.sents):
        n_tokens = len(sentence)
        window = min(n, n_tokens)
        for j in range(n_tokens):
            for k in range(j + 1, min(j + 1 + window, n_tokens + 1)):
                yield sentence_id, sentence[j:k]


def _merge_surface_forms(surface_forms: Iterator[Tuple[int, Span]]) -> Iterable[Candidate]:
    """De-dup candidate surface forms.

    Args:
        surface_forms (Iterable): tuples of <sent_i, span>.

    Returns:
        List of candidates.
    """
    candidates = dict()
    for sent_i, span in surface_forms:
        idx = span.lemma_.lower()
        try:
            c = candidates[idx]
        except (KeyError, IndexError):
            lexical_form = [token.lemma_.lower() for token in span]
            c = candidates[idx] = Candidate(lexical_form, [], [])
        c.surface_forms.append(span)
        c.sentence_ids.append(sent_i)
    return list(candidates.values())



from typing import Any, Dict, List, Tuple

import networkx as nx
from spacy.tokens.doc import Doc

from spacy_ke.base import KeywordExtractor
from spacy_ke.util import Candidate


class PositionRank(KeywordExtractor):
    """spaCy implementation of "PositionRank: An Unsupervised Approach to Keyphrase Extraction from Scholarly".

    Usage example:
    --------------

    >>> import spacy

    >>> nlp = spacy.load("en_core_web_sm")
    >>> nlp.add_pipe(PositionRank(nlp))

    >>> doc = nlp(
        "Natural language processing (NLP) is a subfield of linguistics, computer science, and artificial intelligence "
        "concerned with the interactions between computers and human language, in particular how to program computers "
        "to process and analyze large amounts of natural language data. "
    )

    >>> doc._.extract_keywords(n=5)
    """

    defaults: Dict[str, Any] = {
        "pos": frozenset({"ADJ", "NOUN", "PROPN"}),
        "window": 10,
        "alpha": 0.85,
        "tol": 1.0e-5,
        "normalize": False,
        "candidate_selection": "chunk",
    }

    def candidate_weighting(self, doc: Doc) -> List[Tuple[Candidate, float]]:
        """Compute the weighted score of each keyword candidate.

        Args:
            doc (Doc): doc.

        Returns:
            list of tuples, candidate with a score.
        """
        res = []

        # build the word graph
        G = self.build_graph(doc)

        # normalize cumulated inverse positions
        positions = nx.get_node_attributes(G, "position")
        norm = sum(positions.values())
        for word in positions:
            positions[word] /= norm

        # compute the word scores using biased random walk
        W = nx.pagerank_scipy(
            G,
            alpha=self.cfg["alpha"],
            tol=self.cfg["tol"],
            personalization=positions,
            weight="weight",
        )

        for candidate in doc._.kw_candidates:
            candidate_w = sum([W.get(t, 0.0) for t in candidate.lexical_form])
            if self.cfg["normalize"]:
                candidate_w /= len(candidate.lexical_form)
            res.append((candidate, candidate_w))

        # sort by score
        res.sort(key=lambda x: x[1], reverse=True)
        return res

    def build_graph(self, doc: Doc):
        """Build a lemma graph representation of a document.

        Args:
            doc (Doc): doc.

        Returns:
            nx.Graph
        """
        G = nx.Graph()
        pos = self.cfg["pos"]
        window_size = self.cfg["window"]
        n_tokens = len(doc)

        for token in doc:
            if token.pos_ not in pos:
                continue
            node0 = token.lemma_.lower()
            position0 = token.i
            if not G.has_node(node0):
                G.add_node(node0, position=1 / (position0 + 1))

            j = position0 + 1
            while j < n_tokens and (doc[j].i - token.i) < window_size:
                node1, position1 = doc[j].lemma_.lower(), doc[j].i
                if node0 != node1:
                    if not G.has_node(node1):
                        G.add_node(node1, position=(1 / position1 + 1))
                    if not G.has_edge(node0, node1):
                        G.add_edge(node0, node1, weight=0)
                    G[node0][node1]["weight"] += 1
                j = j + 1

        return G



from spacy.language import Language

from spacy_ke.about import *
from spacy_ke.positionrank import PositionRank
from spacy_ke.textrank import TextRank
from spacy_ke.topicrank import TopicRank
from spacy_ke.yake import Yake

if hasattr(Language, "factory"):
    from thinc.api import Config

    defaults = Config().from_str(
        """
    [Yake]
    window = 2
    lemmatize = false
    candidate_selection = ngram
    
    [TextRank]
    window = 3
    alpha = 0.85
    tol = 1.0e-6
    candidate_selection = chunk
    
    [PositionRank]
    window = 10
    alpha = 0.85
    tol = 1.0e-5
    normalize = false
    candidate_selection = chunk
    
    [TopicRank]
    clustering_method = average
    distance_metric = jaccard
    threshold = 0.74
    alpha = 0.85
    tol = 1.0e-6
    heuristic = null
    candidate_selection = chunk
    """
    )

    @Language.factory("yake", default_config=defaults["Yake"])
    def make_yake(
        nlp: Language, name: str, window: int, lemmatize: bool, candidate_selection: str
    ):
        return Yake(
            nlp,
            window=window,
            lemmatize=lemmatize,
            candidate_selection=candidate_selection,
        )

    @Language.factory("textrank", default_config=defaults["TextRank"])
    def make_textrank(
        nlp: Language,
        name: str,
        window: int,
        alpha: float,
        tol: float,
        candidate_selection: str,
    ):
        return TextRank(
            nlp,
            window=window,
            alpha=alpha,
            tol=tol,
            candidate_selection=candidate_selection,
        )

    @Language.factory("positionrank", default_config=defaults["PositionRank"])
    def make_positionrank(
        nlp: Language,
        name: str,
        window: int,
        alpha: float,
        tol: float,
        normalize: bool,
        candidate_selection: str,
    ):
        return PositionRank(
            nlp,
            window=window,
            alpha=alpha,
            tol=tol,
            normalize=normalize,
            candidate_selection=candidate_selection,
        )

    @Language.factory("topicrank", default_config=defaults["TopicRank"])
    def make_topicrank(
        nlp: Language,
        name: str,
        clustering_method: str,
        distance_metric: str,
        threshold: float,
        alpha: float,
        tol: float,
        heuristic: str,
        candidate_selection: str,
    ):
        return TopicRank(
            nlp,
            clustering_method=clustering_method,
            distance_metric=distance_metric,
            threshold=threshold,
            alpha=alpha,
            tol=tol,
            heuristic=heuristic,
            candidate_selection=candidate_selection,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# inspired from:
# https://github.com/kennethreitz/setup.py/blob/master/setup.py
# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine setuptools --dev

import io
import os
import sys
from shutil import rmtree
from setuptools import find_packages, setup, Command

root = os.path.abspath(os.path.dirname(__file__))
sys.path.append(root)

with io.open(os.path.join(root, "spacy_ke", "about.py"), encoding="utf8") as f:
    about = {}
    exec(f.read(), about)

# Use the README and use it as the long-description.
with io.open(os.path.join(root, "README.md"), encoding="utf-8") as f:
    long_description = "\n" + f.read()

with io.open(os.path.join(root, "requirements.txt"), encoding="utf-8") as f:
    install_requires = [line.strip() for line in f.readlines() if not line.startswith("spacy")]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(root, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")
        sys.exit()


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__summary__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__url__"],
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    install_requires=install_requires,
    python_requires=">=3.6",
    include_package_data=True,
    license=about["__license__"],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    cmdclass={
        "upload": UploadCommand,
    },
    entry_points={}
)
